######################
## Main VAGAM main fitting function
## smooth.X and para.X correspond to the covariates entered as smooth terms and parameteric terms respectively
## int.knots is the number of interior knots to use for each covariate. Equally spaced knots based on quantiles are used
## control$mc.samps controls Monte Carlo samples for calculating variational observed information matrix using Louis' method
## control$seed controls seed for starting values for Poisson and the seed of the fitting algorithm in general, which can be a tad unstable
## control$pois.step.size controls step size for Poisson P-IRLS, which can be a tad unstable
####################

# library(splines)
# library(mgcv)
# library(gamm4)
library(Matrix)
library(mvtnorm)
library(truncnorm)

vagam <- function(y, smooth.X, para.X = NULL, lambda = NULL, int.knots, family = gaussian(), A.struct = c("unstructured", "block"), save.data = FALSE, para.se = FALSE, doAIC = FALSE, control = list(eps = 1e-3, maxit = 1000, trace = TRUE, seed.number = 123, mc.samps = 4000, pois.step.size = 0.01))
{  
	if(!(family$family %in% c("gaussian", "poisson", "binomial","binomiall")))
		stop("Specified family currently no supported. Sorry!")
 	if(family$family == "binomial")
 		family <- binomial(link = "probit")
		
	A.struct <- match.arg(A.struct, c("unstructured", "block"))
	ncov.smooth.X <- ncol(smooth.X)
		
	if(!(length(int.knots) %in% c(1,ncov.smooth.X))) 
		stop("int.knots should either be a single number of a vector with length equal to ncol(smooth.X). Thanks!")
	if(length(int.knots) == 1) 
		actual.int.knots <- rep(int.knots, ncov.smooth.X)
	if(length(int.knots) > 1) 
		actual.int.knots <- int.knots

	if(is.null(colnames(smooth.X))) colnames(smooth.X) <- paste("smoothX", 1:ncov.smooth.X, sep = "")
	if(!is.null(para.X))
	{
		para.X <- as.matrix(para.X)
		if(is.null(colnames(para.X))) colnames(para.X) <- paste("paraX", 1:ncol(para.X), sep = "")
		if(any(apply(para.X, 2, function(x) all(x == 1))))
			stop("No intercept terms should be included in para.X, as this is included by default. Thanks!")
		para.X <- cbind(1, para.X) ## Intercept included by default
		colnames(para.X)[1] <- "Intercept"
	}
	if(is.null(para.X)) 
	{
		para.X <- matrix(1, nrow = length(y), ncol = 1)
		colnames(para.X) <- "Intercept"
	}
	if(is.null(lambda))
	{
		cw.lambda <- new.lambda <- rep(2, ncov.smooth.X)
		message("Lambda updated as part of VA estimation. Yeah baby!")
	}
	if(!is.null(lambda)) 
	{
		if(length(lambda) != ncov.smooth.X)
			stop("lambda should a vector with length equal to ncol(smooth.X). Thanks!")
		cw.lambda <- new.lambda <- lambda 
		message("Lambda given. Thanks buddy!")
	}

	
	## Construct B-splines basis, penalty matrix, and some other quantities for smooth.X, by extracting attributes from a mgcv gam fit
     ## no. of total knots = no. of interior knots[1] + 2*degree + 1, but then -1 to remove the intercept term 
     ## no. of basis dimension = d_j = no. of interior knots[1]+ degree + 1, where degree of spline is typically set to 3 for cubic B-splines. But then we typically -1 to remove the intercept term. Then to ensure centering constraint is the satisfied, reduce by one to produce to d_j = no. of interior knots[1] + degree - 1 
	S <- basis.info <- vector("list", ncov.smooth.X) 
	d <- numeric(ncov.smooth.X) ## rank of Sj
	index.cov <- NULL ## Indexes which covariate each column in Z belongs to
	Z <- NULL

	for(k in 1:ncov.smooth.X)
	{
 		get.basis <- smooth.construct(s(x0, bs = "ps", k = actual.int.knots[k]+3, m  = c(2,1)), data = data.frame(x0 = smooth.X[,k]), knots = NULL)
 		Z.k <- get.basis$X
 		transform.mat <- qr.Q(qr(as.matrix(colSums(Z.k))), complete = TRUE)[,-1] ## Centering constraint
		Z.k <- Z.k %*% transform.mat
		index.cov <- c(index.cov, rep(k, ncol(Z.k)))
		colnames(Z.k) <- paste("smooth.X", k, 1:sum(index.cov == k), sep = "")
		Z <- cbind(Z, Z.k)
 		S[[k]] <- t(transform.mat) %*% get.basis$S[[1]] %*% transform.mat
 		d[k] <- get.basis$rank
 		get.basis$X <- NULL; 
 		get.basis$transform.mat <- transform.mat; 
 		basis.info[[k]] <- get.basis
 		rm(transform.mat, get.basis)
	}

	
	cw.kappa <- new.kappa <- numeric(ncol(para.X)) ## coefficients for parametric part
	names(new.kappa) <- names(para.X)
	cw.a <- new.a <- numeric(ncol(Z)) ## coefficients for smooth part, also VA means
	if(family$family == "poisson") 
	{
		old <- .Random.seed
		on.exit({ .Random.seed <<- old})
		set.seed(control$seed.number)
		cw.a <- new.a <- rnorm(ncol(Z)) ## Use random coefficients as it tends to work better for Poisson
	}
	names(new.a) <- colnames(Z)
	cw.A <- new.A <- matrix(0, nrow = ncol(Z), ncol = ncol(Z)) ## VA covariance
	rownames(new.A) <- colnames(new.A) <- colnames(Z)
	cw.phi <- new.phi <- 1 ## variance
	cw.logL <- -Inf; new.logL <- 10; diff.logL <- 10; diff.lambda <- 0; counter <- 1
		
	if(family$family %in% c("gaussian","binomial")) ZZt <- crossprod(Z)
	tic <- proc.time()

	## Let it rip!
	while(diff.logL > control$eps)
	{ 
		if(counter > control$maxit) 
               break;
	
		## Update kappa and a and lambda if required
		## Calculation for Poisson somewhat unstable
		ZA <- Z %*% cw.A
		new.offset <- Z %*% cw.a + as.numeric(family$family %in% c("poisson","binomiall")) * 0.5 * rowSums(ZA * Z)
 		fit0 <- suppressWarnings(glm.fit(x = para.X, y = y, family = family, offset = new.offset, intercept = FALSE))
 		new.kappa <- fit0$coefficients
 		
		Q <- matrix(0, ncol(Z), ncol(Z))
		for(k2 in 1:ncov.smooth.X) Q[index.cov == k2,index.cov == k2] <- S[[k2]]*cw.lambda[k2] 
		new.offset <- para.X %*% new.kappa + as.numeric(family$family %in% c("poisson","binomiall")) * 0.5 * rowSums(ZA * Z)
		if(family$family == "poisson")
          {
 			fit0 <- ridge.iwls(x = Z, y = y, family = family, penalty = Q, initial.beta = cw.a, offset = new.offset, gamma = 0.1 + min(counter * control$pois.step.size, 0.9), extra = doAIC) ## Gradually increase step size with counter
		}
		if(family$family != "poisson")
		{
			fit0 <- ridge.iwls(x = Z, y = y, family = family, penalty = cw.phi*Q, initial.beta = cw.a, offset = new.offset, extra = doAIC) 
		}
 		new.a <- fit0$coefficients
		rm(new.offset)
		
		
		## Update phi for Gaussian
		new.eta <- para.X %*% new.kappa + Z %*% new.a
		if(family$family == "gaussian") 
		{
			new.phi <- (sum((y - new.eta)^2) + sum(ZA * Z))/length(y)
		}
		
		
		## Update A
		if(A.struct == "unstructured") 
		{
			if(family$family %in% c("gaussian","binomial"))
				new.A <- chol2inv(chol(Q + (1/new.phi)*ZZt))
			if(family$family == "poisson")
			{
				err2 <- 10; 
				cw.A <- new.A
				while(err2 > 0.01) 
				{
					denom <- Z * matrix(sqrt(exp(new.eta + 0.5 * rowSums(ZA * Z))), nrow = length(y), ncol = ncol(Z), byrow = FALSE)
					denom <- Q + crossprod(denom)
					new.A <- chol2inv(chol(denom))
					err2 <- sum((cw.A - new.A)^2)/2
					cw.A <- new.A
				}
			}
			if(family$family == "binomiall")
			{
				err2 <- 10; 
				cw.A <- new.A
				while(err2 > 0.01) 
				{
					denom <- Z * matrix(sqrt(binomial()$linkinv(new.eta + 0.5 * rowSums(ZA * Z))), nrow = length(y), ncol = ncol(Z), byrow = FALSE)
					denom <- Q + crossprod(denom)
					new.A <- chol2inv(chol(denom))
					err2 <- sum((cw.A - new.A)^2)/2
					cw.A <- new.A
				}
			}
		}
		if(A.struct == "block") ## This might actually be slower than doing the full thing!
		{
			for(k2 in 1:ncov.smooth.X) 
			{
				if(family$family %in% c("gaussian", "binomial"))
				{
					new.A[index.cov == k2, index.cov == k2] <- chol2inv(chol(Q[index.cov == k2,index.cov == k2] + (1/new.phi) * crossprod(Z[,index.cov == k2])))
 				}
				if(family$family == "poisson")
				{
					err2 <- 10; 
					while(err2 > 0.01) 
						{
						denom <- Z[,index.cov == k2] * matrix(sqrt(exp(new.eta + 0.5 * rowSums(ZA * Z))), nrow = length(y), ncol = sum(index.cov==k2), byrow = FALSE)
						denom <- Q[index.cov == k2, index.cov == k2] + crossprod(denom)
						new.A[index.cov == k2, index.cov == k2] <- chol2inv(chol(denom))
						err2 <- sum((cw.A - new.A)^2)/2
						cw.A <- new.A
						}
				}
				if(family$family == "binomiall")
				{
					err2 <- 10; 
					while(err2 > 0.01) 
						{
						denom <- Z[,index.cov == k2] * matrix(sqrt(binomial()$linkinv(new.eta + 0.5 * rowSums(ZA * Z))), nrow = length(y), ncol = sum(index.cov==k2), byrow = FALSE)
						denom <- Q[index.cov == k2, index.cov == k2] + crossprod(denom)
						new.A[index.cov == k2, index.cov == k2] <- chol2inv(chol(denom))
						err2 <- sum((cw.A - new.A)^2)/2
						cw.A <- new.A
						}
				}
			}
		}	

		if(is.null(lambda))
		{
			for(k2 in 1:ncov.smooth.X) 
			{
				new.lambda[k2] <- d[k2]/(sum(diag(crossprod(S[[k2]], new.A[index.cov == k2, index.cov == k2]))) + tcrossprod(new.a[index.cov == k2], S[[k2]]) %*% new.a[index.cov == k2])
			}
		}

		
		## Calculate new VA logL
		new.logL <- calc.VAlogL(y = y, Z = Z, para.X = para.X, family = family, lambda = new.lambda, kappa = new.kappa, a = new.a, A = new.A, phi = new.phi, S = S, d = d, index.cov = index.cov, eta = new.eta)
		diff.logL <- new.logL - cw.logL
		diff.lambda <- sum((new.lambda - cw.lambda)^2)
		if(control$trace) 
		{
			cat("Iteration:", counter, "\t Current VA logL:", cw.logL, " | New VA logL:", new.logL, " | Difference:", new.logL - cw.logL, "\n")
		}
		counter <- counter + 1
		cw.logL <- new.logL
		cw.lambda <- new.lambda
		cw.kappa <- new.kappa
		cw.a <- new.a
		cw.A <- new.A
		cw.phi <- new.phi
		}
	toc <- proc.time()

	
	## Calculate stuff related to external smoothing parameter selection
     aic.out <- NA
	if(doAIC)    
          aic.out <- -2 * new.logL - new.phi + fit0$tr.Inf * new.phi

	out <- list(kappa = new.kappa, a = new.a, A = new.A, lambda = new.lambda, AIC = aic.out, phi = new.phi, linear.predictors = new.eta, logL = new.logL, no.knots = actual.int.knots, index.cov = index.cov, basis.info = basis.info, family = family, time.taken = toc-tic)
	if(save.data) 
	{
		out$y
		out$para.X <- para.X
		out$smooth.X <- smooth.X
		out$Z <- Z
	}

	
	## Wald test for each smooth curve
	smooth.X.waldstat <- smooth.X.waldpval <- numeric(ncov.smooth.X)
 	new.A.inv <- chol2inv(chol(new.A))
	for(k2 in 1:ncov.smooth.X) 
	{
		smooth.X.waldstat[k2] <- t(new.a[index.cov == k2])%*%new.A.inv[index.cov == k2,index.cov == k2]%*%new.a[index.cov == k2]
		smooth.X.waldpval[k2] <- pchisq(smooth.X.waldstat[k2], df = sum(index.cov==k2), lower = FALSE)
	}
	smooth.X.wald <- round(rbind(smooth.X.waldstat,smooth.X.waldpval),5)
	rownames(smooth.X.wald) <- c("Wald Statistic", "p-value")
	colnames(smooth.X.wald) <- colnames(smooth.X)
	rm(smooth.X.waldstat,smooth.X.waldpval)
	out$smooth.stat <- smooth.X.wald
	
	
	## Variatonal Observed information matrix for model params and lambas
	if(para.se) 
	{
		message("Calculating information matrix for model parameters...")
 		obs.info <- info.valouis(y = y, para.X = para.X, Z = Z, kappa = new.kappa, phi = new.phi, lambda = new.lambda, a = new.a, A = new.A, S = S, d = d, index.cov = index.cov, mc.samps = control$mc.samps, family = family)                                 
		## If there the information matrix is not +ve definite, then do a one-step to produce an unstructured A and recalculate the info matrix; hopefully this helps!
		if(any(diag(obs.info) < 0) & family$family == "poisson") 
		{
		message("Redoing...")
		err2 <- 10; 
		cw.A <- new.A
		#new.eta <- para.X %*% new.kappa + Z %*% new.a
		while(err2 > 0.01) 
			{
			denom <- Z * matrix(sqrt(exp(new.eta + 0.5 * rowSums((Z %*% cw.A) * Z))), nrow = length(y), ncol = ncol(Z), byrow = FALSE)
			denom <- Q + crossprod(denom)
			new.A <- chol2inv(chol(denom))
			err2 <- sum((cw.A - new.A)^2)/2
			cw.A <- new.A
			}
		obs.info <- info.valouis(y = y, para.X = para.X, Z = Z, kappa = new.kappa, phi = new.phi, lambda = new.lambda, a = new.a, A = cw.A, S = S, d = d, index.cov = index.cov, mc.samps = control$mc.samps, family = family)                                 
		}

 		while(class(try(solve(obs.info),silent=TRUE)) == "try-error")
			{
			obs.info <- info.valouis(y = y, para.X = para.X, Z = Z, kappa = new.kappa, phi = new.phi, lambda = new.lambda, a = new.a, A = new.A, S = S, d = d, index.cov = index.cov, mc.samps = control$mc.samps, family = family)
			}
		while(!all(diag(solve(obs.info)) > 0))
			{
			obs.info <- info.valouis(y = y, para.X = para.X, Z = Z, kappa = new.kappa, phi = new.phi, lambda = new.lambda, a = new.a, A = new.A, S = S, d = d, index.cov = index.cov, mc.samps = control$mc.samps, family = family)
			while(class(try(solve(obs.info),silent=TRUE)) == "try-error")
				{
				obs.info <- info.valouis(y = y, para.X = para.X, Z = Z, kappa = new.kappa, phi = new.phi, lambda = new.lambda, a = new.a, A = new.A, S = S, d = d, index.cov = index.cov, mc.samps = control$mc.samps, family = family)
				}
			}
		para.sderr <- sqrt(diag(solve(obs.info)))[1:ncol(para.X)]
		para.X.stat <- round(rbind(new.kappa, para.sderr, new.kappa/para.sderr, 2 * pnorm(abs(new.kappa/para.sderr), lower = FALSE)), 5)
		rownames(para.X.stat) <- c("Estimate", "Std. Error", "Wald Statistic", "p-value")
		colnames(para.X.stat) <- colnames(para.X)
		out$para.stat <- para.X.stat
		out$obs.info <- obs.info
		}
		
	class(out) <- "vagam"
	out$call <- match.call()
	return(out)	
}
   	   

   	  